{
    Super Nova effect DLL for Borland Delphi version 5 and later.
    Version 1.00 (build: 2002/09/30)
    Copyright (C) 2002 MASAHIRO SUZUKI <suzuki@woodybells.com>

    Original source code is GIMP(www.gimp.org) SuperNova plug-in.

    * SuperNova plug-in
    * Copyright (C) 1997 Eiichi Takamori <taka@ma1.seikyou.ne.jp>,
    *                    Spencer Kimball, Federico Mena Quintero

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
library Nova;

uses
  Windows, Graphics;

type
  TProgressProc   = procedure(Sender: TObject; Percentdone: Integer) of object;
  PRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array [0..0] of TRGBTriple;
  PDoubleArray    = ^TDoubleArray;
  TDoubleArray    = array [0..0] of Double;
  PIntegerArray   = ^TIntegerArray;
  TIntegerArray   = array [0..0] of Integer;

function Max(A,B: Double): Double;
begin
  if A > B then
    Result := A
  else
    Result := B;
end;

function Min(A,B: Double): Double;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

function ArcTan2(Y, X: Extended): Extended;
asm
  FLD     Y
  FLD     X
  FPATAN
  FWAIT
end;

function Floor(X: Extended): Integer;
begin
  Result := Integer(Trunc(X));
  if Frac(X) < 0 then
    Dec(Result);
end;

function RoundByte(Data: Double): Byte;
begin
  if Data > 255 then Data := 255;
  if Data <   0 then Data := 0;
  Result := Trunc(Data + 0.5);
end;

function Gauss: Double;
var
  Sum: Double;
  i: Integer;
begin
  Sum:= 0;
  for i := 1 to 6 do
    Sum := Sum + (Random(2147483647) / 2147483647);
    Result := Sum / 6;
end;

function Clamp(i, l, h: Double): Double;
begin
  if i < l then Result := l
  else if i > h then Result := h
  else Result := i;
end;

procedure RGB2HSV(R, G, B: Double; var H, S, V: Double);
var
  vv, m, vm: Double;
  r2, g2, b2: Double;
begin
  H := 0; S := 0; V := 0;
  vv := Max(Max(R, G), B);
  m := Min(Min(R, G), B);
  V := (m + vv) / 2;
  if V <= 0 then exit;
  vm := vv - m;
  s  := vm;
  if s > 0 then
  begin
    if V <= 0.5 then
      S := S / (vv + m)
    else S := S / (2 - vv - m);
  end else exit;
  r2 := (vv - r) / vm;
  g2 := (vv - g) / vm;
  b2 := (vv - b) / vm;
  if r = vv then
    if G = m then H := b2 + 5
    else          H := 1 - g2
  else if G = vv then
    if b = m then H := 1 + r2
    else          H := 3 - b2
  else
    if R = m then H := 3 + g2
    else          H := 5 - r2;
  H := H / 6;
end;

procedure HSV2RGB(H, S, V: Double; var R, G, B: Double);
var
  vv, m, sv, vsf: Double;
  sextant: Integer;
begin
  if H = 1.0 then H := 0.0;
  if V <= 0.5 then
    vv := V * (1 + S)
  else vv := V + S - V * S;
  if vv <= 0 then
  begin
    R := 0; G := 0; B := 0;
  end else
  begin
    m := V + V - vv;
    sv := (vv - m) / vv;
    H := H * 6.0;
    sextant := Trunc(H);
    vsf := vv * sv * (H - sextant);
    case sextant of
      0: begin
           R := vv; G := m + vsf; B := m;
         end;
      1: begin
           R := vv - vsf; G := vv; B := m;
         end;
      2: begin
           R := m; G := vv; B := m + vsf;
         end;
      3: begin
           R := m; G := vv - vsf; B := vv;
         end;
      4: begin
           R := m + vsf; G := m; B := vv;
         end;
      5: begin
           R := vv; G := m; B := vv - vsf;
         end;
      end;
   end;
end;

function SuperNova(Bmp: TBitmap; AColor: TColor; cx, cy, Radius, Spokes,
                    RandomHue: Integer; Progress: TProgressProc): Boolean;
var
  x, y, BmpW, BmpH: Integer;
  l, u, v, w, w1, c,
  SpokeR, SpokeG, SpokeB,
  NovaAlpha, ComplRatio, Ratio,
  rgb_r, rgb_g, rgb_b,
  hsv_h, hsv_s, hsv_v: Double;
  Spoke: PDoubleArray;
  SpokeColor: PIntegerArray;
  i: Integer;
  pLine: PRGBTripleArray;
begin
  Result := False;
  if Bmp.Empty then Exit;
  Bmp.PixelFormat := pf24bit;
  BmpW := Bmp.Width;
  BmpH := Bmp.Height;
  Randomize;
  if Spokes < 1 then Spokes := 1
  else if Spokes > 2000 then Spokes := 2000;
  if Radius < 1 then Radius := 1
  else if Radius > 200 then Radius := 200;
  if RandomHue < 0 then RandomHue := 0
  else if RandomHue > 360 then RandomHue := 360;
  GetMem(Spoke, Spokes * SizeOf(Double));
  GetMem(SpokeColor, Spokes * SizeOf(Integer) * 3);
  try
    RGB2HSV(GetRValue(ColorToRgb(AColor)) / 255.0,
            GetGValue(ColorToRgb(AColor)) / 255.0,
            GetBValue(ColorToRgb(AColor)) / 255.0, hsv_h, hsv_s, hsv_v);
    for i := 0 to Spokes - 1 do
    begin
      Spoke[i] := Gauss;
      hsv_h := hsv_h + RandomHue / 360.0 *
               (Random(2147483647) / 2147483647 - 0.5);
      if hsv_h < 0 then
        hsv_h := hsv_h + 1.0
      else if hsv_h > 1.0 then
        hsv_h := hsv_h - 1.0;
      HSV2RGB(hsv_h, hsv_s, hsv_v, rgb_r, rgb_g, rgb_b);
      SpokeColor[3 * i + 0] := Trunc(255 * rgb_r);
      SpokeColor[3 * i + 1] := Trunc(255 * rgb_g);
      SpokeColor[3 * i + 2] := Trunc(255 * rgb_b);
    end;
    for y := 0 to BmpH - 1 do
    begin
      pLine := Bmp.ScanLine[y];
      for x := 0 to BmpW - 1 do
      begin
        u := (y - cy) / Radius;
        v := (x - cx) / Radius;
        l := Sqrt((u * u) + (v * v));
        c := (ArcTan2(u, v) / (2 * Pi) + 0.51) * Spokes;
        i := Floor(c);
        c := c - i;
        i := i mod Spokes;
        w1 := Spoke[i] * (1 - c) + Spoke[(i + 1) mod Spokes] * c;
        w1 := w1 * w1;
        w := 1 / (l + 0.001) * 0.9;
        NovaAlpha := Clamp(w, 0.0, 1.0);
        Ratio := NovaAlpha;
        ComplRatio := 1.0 - Ratio;
        SpokeR := SpokeColor[3 * i + 0] * (1 - c) +
                  SpokeColor[3 * ((i + 1) mod Spokes) + 0] * c;
        SpokeG := SpokeColor[3 * i + 1] * (1 - c) +
                  SpokeColor[3 * ((i + 1) mod Spokes) + 1] * c;
        SpokeB := SpokeColor[3 * i + 2] * (1 - c) +
                  SpokeColor[3 * ((i + 1) mod Spokes) + 2] * c;
        c := Clamp(w1 * w, 0.0, 1.0) * 255;
        with pLine^[x] do
        begin
          if w > 1.0 then
          begin
            rgbtRed   := RoundByte(SpokeR * w + c);
            rgbtGreen := RoundByte(SpokeG * w + c);
            rgbtBlue  := RoundByte(SpokeB * w + c);
          end else
          begin
            rgbtRed   := RoundByte((rgbtRed   * ComplRatio + SpokeR * Ratio) + c);
            rgbtGreen := RoundByte((rgbtGreen * ComplRatio + SpokeG * Ratio) + c);
            rgbtBlue  := RoundByte((rgbtBlue  * ComplRatio + SpokeB * Ratio) + c);
          end;
        end;
      end;
      if Assigned(Progress) then
        Progress(nil, (y + 1) * 100 div BmpH);
    end;
    Result := True;
  finally
    FreeMem(Spoke);
    FreeMem(SpokeColor);
  end;
end;

exports SuperNova;

begin
end.
